__package_name__: str = 'python-utils'
__author__: str = 'Rick van Hattem'
__author_email__: str = 'Wolph@wol.ph'
__description__: str = (
    'Python Utils is a module with some convenient utilities not included '
    'with the standard Python install'
)
__url__: str = 'https://github.com/WoLpH/python-utils'
# Omit type info due to automatic versioning script
__version__ = '3.3.3'
